/****************************************************************************
*   Squeak port for OS/2
*
*   - MIDI primitives
*
* THIS IS UNDER CONSTRUCTION - PLEASE FORGIVE THE COMMENTS IN SPANISH
*     (They won't be in the final version!)
*
* Feel free to experiment with this files, and to modify them anyway you like.
* If you fix any bugs, or enhance something, please let me know. I will
* include any enhancement on future releases. Bug reports are also welcome.
* I compile Squeak with IBM's VisualAge C++ for OS/2, but I think you can
* use other compilers as well if they support multimedia APIs (MMPM2.LIB).
* Enyoy!
*
*   Juan Manuel Vuletich
*    jvuletic@dc.uba.ar
*
*****************************************************************************/
/*
   Currently we support the MPUDev driver, and just one real MPU compatible
   MIDI port. It is a bidirectional device.
*/
/* JMV

1ra etapa: algo elemental que funcione.
Quizas incluyendo un thread de TC que haga polling del port.
Esto mejora el timestamping, pero tiene el problema de que puede dilatar
aun mas la sintesis en tiempo real.

Esta version solo soporta un port MPU simple. Deberia tambien soportar o el
sequencer OS/2 o alguna forma de acceder al sintetizador de la placa de sonido
*/

#define  INCL_DOSPROCESS            /* Para procesos y threads */
#include <os2.h>
#include "sq.h"
#include "MIDIPlugin.h"

#ifndef NO_MIDI


/*** MIDI Parameters (used with sqMIDIParameter function) ***/

#define sqMIDIInstalled            1
/* Read-only. Return 1 if a MIDI driver is installed, 0 if not.
   On OMS-based MIDI drivers, this returns 1 only if the OMS
   system is properly installed and configured. */

#define sqMIDIVersion            2
/* Read-only. Return the integer version number of this MIDI driver.
   The version numbering sequence is relative to a particular driver.
   That is, version 3 of the Macintosh MIDI driver is not necessarily
   related to version 3 of the Win95 MIDI driver. */

#define sqMIDIHasBuffer            3
/* Read-only. Return 1 if this MIDI driver has a time-stamped output
   buffer, 0 otherwise. Such a buffer allows the client to schedule
   MIDI output packets to be sent later. This can allow more precise
   timing, since the driver uses timer interrupts to send the data
   at the right time even if the processor is in the midst of a
   long-running Squeak primitive or is running some other application
   or system task. */

#define sqMIDIHasDurs            4
/* Read-only. Return 1 if this MIDI driver supports an extended
   primitive for note-playing that includes the note duration and
   schedules both the note-on and the note-off messages in the
   driver. Otherwise, return 0. */

#define sqMIDICanSetClock         5
/* Read-only. Return 1 if this MIDI driver''s clock can be set
   via an extended primitive, 0 if not.*/

#define sqMIDICanUseSemaphore      6
/* Read-only. Return 1 if this MIDI driver can signal a semaphore
   when MIDI input arrives. Otherwise, return 0. If this driver
   supports controller caching and it is enabled, then incoming
   controller messages will not signal the semaphore. */

#define sqMIDIEchoOn            7
/* Read-write. If this flag is set to a non-zero value, and if
   the driver supports echoing, then incoming MIDI events will
   be echoed immediately. If this driver does not support echoing,
   then queries of this parameter will always return 0 and
   attempts to change its value will do nothing. */

#define sqMIDIUseControllerCache   8
/* Read-write. If this flag is set to a non-zero value, and if
   the driver supports a controller cache, then the driver will
   maintain a cache of the latest value seen for each MIDI controller,
   and control update messages will be filtered out of the incoming
   MIDI stream. An extended MIDI primitive allows the client to
   poll the driver for the current value of each controller. If
   this driver does not support a controller cache, then queries
   of this parameter will always return 0 and attempts to change
   its value will do nothing. */

#define sqMIDIEventsAvailable      9
/* Read-only. Return the number of MIDI packets in the input queue. */

#define sqMIDIFlushDriver         10
/* Write-only. Setting this parameter to any value forces the driver
   to flush its I/0 buffer, discarding all unprocessed data. Reading
   this parameter returns 0. Setting this parameter will do nothing
   if the driver does not support buffer flushing. */

#define sqMIDIClockTicksPerSec      11
/* Read-only. Return the MIDI clock rate in ticks per second. */

#define sqMIDIHasInputClock         12
/* Read-only. Return 1 if this MIDI driver timestamps incoming
   MIDI data with the current value of the MIDI clock, 0 otherwise.
   If the driver does not support such timestamping, then the
   client must read input data frequently and provide its own
   timestamping. */

#define MIDI_IN_DEVICE 1
#define MIDI_OUT_DEVICE 2

#define QSIZE 50
char InputQ[QSIZE];
int WritingPos = 0;
int ReadingPos = 0;
TID tidInputThread;
HFILE mpuHnd = 0;

int SqueakSemaphore=0;

int sqMIDIClosePort(int portNum);

/* This Time Critical thread wakes when there is MIDI input.
   Then, it puts the data in the queue and signals the
   Squeak semaphore */
VOID _System InputThreadFunc(ULONG parameter) {
   ULONG bytesRead;
   ULONG bytesToRead;
   while(1) {
      if(WritingPos < ReadingPos)
         bytesToRead = ReadingPos - WritingPos;
      else
         bytesToRead = QSIZE - WritingPos;
      DosRead(mpuHnd, &(InputQ[WritingPos]), bytesToRead, &bytesRead);
      WritingPos = WritingPos + bytesRead;
      if(WritingPos == QSIZE)
         WritingPos = 0;
      if(SqueakSemaphore)
         signalSemaphoreWithIndex(SqueakSemaphore);
   }
}

/* Open the given port, if possible. If non-zero, readSemaphoreIndex
   specifies the index in the external objects array of a semaphore
   to be signalled when incoming MIDI data is available. Note that
   not all implementations support read semaphores (this one does);
   see sqMIDICanUseSemaphore. The interfaceClockRate parameter
   specifies the clock speed for an external MIDI interface
   adaptor on platforms that use such adaptors (e.g., Macintosh)
   and is ignored with OS/2.
   Fail if there is no port of the given number.*/
int sqMIDIOpenPort(int portNum, int readSemaIndex, int interfaceClockRate) {
   ULONG ulAction;
   if(portNum < 0 || portNum >= 1)
      return success(false);
   SqueakSemaphore = readSemaIndex;
   /* If asked to open twice, only take the semaphore */
   if(!mpuHnd) {
      DosOpen("MPUDEV1$", &mpuHnd, &ulAction, 0L, FILE_NORMAL, OPEN_ACTION_OPEN_IF_EXISTS, OPEN_ACCESS_READWRITE | OPEN_SHARE_DENYREADWRITE, 0L);
      DosCreateThread(&tidInputThread, &InputThreadFunc, 0, 0, 4096*16);
      DosSetPriority(PRTYS_THREAD, PRTYC_TIMECRITICAL, 0, tidInputThread);
   }
   return 1;
}

/* bufferPtr is the address of the first byte of a Smalltalk
   ByteArray of the given length. Copy up to (length - 4) bytes
   of incoming MIDI data into that buffer, preceded by a 4-byte
   timestamp in the units of the MIDI clock, most significant byte
   first. Implementations that do not support timestamping of
   incoming data as it arrives (see sqMIDIHasInputClock) simply
   set the timestamp to the value of the MIDI clock when this
   function is called. Return the total number of bytes read,
   including the timestamp bytes. Return zero if no data is
   available. Fail if the buffer is shorter than five bytes,
   since there must be enough room for the timestamp plus at
   least one data byte. */
int sqMIDIPortReadInto(int portNum, int count, int bufferPtr) {

   unsigned char *dstPtr;
   ULONG stamp;
   int bytesRead=4;
   int writingPosOnEnter;
   writingPosOnEnter = WritingPos;

   if (count < 5) return success(false);
   if(portNum < 0 || portNum >= 1)
      return success(false);
   dstPtr = (char *)bufferPtr;
   /* jmv Warning: This clock's resolution is just 32 mS. Use HiResTimer! */
   stamp = ioMicroMSecs();
   *(dstPtr++) = (stamp >> 24) & 0xFF;
   *(dstPtr++) = (stamp >> 16) & 0xFF;
   *(dstPtr++) = (stamp >> 8) & 0xFF;
   *(dstPtr++) = (stamp >> 0) & 0xFF;
   while ((ReadingPos != writingPosOnEnter) && (bytesRead <= count)) {
      *(dstPtr++) = InputQ[ReadingPos++];
      bytesRead++;
      if(ReadingPos == QSIZE)
         ReadingPos = 0;
   }
   /* return zero if only the stamp is in the buffer */
   return bytesRead == 4 ? 0 : bytesRead;
}

/* bufferPtr is the address of the first byte of a Smalltalk
   ByteArray of the given length. Send its contents to the given
   port when the MIDI clock reaches the given time. If time equals
   zero, then send the data immediately. Implementations that do
   not support a timestamped output queue, such as this one, always
   send the data immediately; see sqMIDIHasBuffer. */
int sqMIDIPortWriteFromAt(int portNum, int count, int bufferPtr, int time) {
   ULONG bytesWritten;

   if(portNum < 0 || portNum >= 1)
      return success(false);
   DosWrite(mpuHnd, (char*) bufferPtr, count, &bytesWritten);
   return bytesWritten;
}

/* Close the given MIDI port. Do nothing if the port is not open.
   Fail if there is no port of the given number.*/
int sqMIDIClosePort(int portNum) {
   if(portNum < 0 || portNum >= 1)
      return success(false);
   /* Do nothing wrong if asked to close twice */
   if (mpuHnd) {
      DosKillThread(tidInputThread);
      tidInputThread = 0;
      DosClose(mpuHnd);
      mpuHnd = 0;
   }
   SqueakSemaphore = 0;
   return 1;
}

int midiInit(void) {
   return 1;
}

int midiShutdown(void) {
   return 1;
}

/* Return the current value of the clock used to schedule MIDI events.
   The MIDI clock is assumed to wrap at or before half the maximum
   positive SmallInteger value. This allows events to be scheduled
   into the future without overflowing into LargePositiveIntegers.
   This implementation does not support event scheduling, so it
   just returns the value of the Squeak millisecond clock. */
/* jmv Warning: This clock's resolution is just 32 mS. Use HiResTimer! */
int sqMIDIGetClock(void) {
   return ioMicroMSecs();
}

/* Return the number of available MIDI interfaces, including both
   hardware ports and software entities that act like ports. Ports
   are numbered from 0 to N-1, where N is the number returned by this
   primitive. */
int sqMIDIGetPortCount(void) {
   return 1;
}

/* Return an integer indicating the directionality of the given
   port where: 1 = input, 2 = output, 3 = bidirectional. Fail if
   there is no port of the given number. */
int sqMIDIGetPortDirectionality(int portNum) {
   if(portNum < 0 || portNum >= 1)
      return success(false);
   return MIDI_OUT_DEVICE | MIDI_IN_DEVICE;
}

/* Copy the name of the given MIDI port into the string at the given
   address. Copy at most length characters, and return the number of
   characters copied. Fail if there is no port of the given number.*/
int sqMIDIGetPortName(int portNum, int namePtr, int length) {
   int count, i;
   char portName[] = "MPU-401";

   if(portNum < 0 || portNum >= 1)
      return success(false);
   count = strlen(portName);
   if (count > length) count = length;
   for(i=0;i<count;i++)
      ((char*)namePtr)[i] = (char) portName[i];
   return count;
}

/* Read or write the given MIDI driver parameter. If modify is 0,
   then newValue is ignored and the current value of the specified
   parameter is returned. If modify is non-zero, then the specified
   parameter is set to newValue. Note that many MIDI driver parameters
   are read-only; attempting to set one of these parameters fails.
   For boolean parameters, true = 1, false = 0. */
int sqMIDIParameter(int whichParameter, int modify, int newValue) {
  int i;

   if (modify == 0) {
      switch(whichParameter) {
      case sqMIDIInstalled:          return 1;
      case sqMIDIVersion:            return 100;
      case sqMIDIHasBuffer:          return 0;
      case sqMIDIHasDurs:            return 0;
      case sqMIDICanSetClock:        return 0;
      case sqMIDICanUseSemaphore:    return 1;
      case sqMIDIEchoOn:             return 0;
      case sqMIDIUseControllerCache: return 0;
      case sqMIDIEventsAvailable:
         if(WritingPos != ReadingPos)
            return 1;
         return 0;
      case sqMIDIFlushDriver:        return 0;
      case sqMIDIClockTicksPerSec:   return 1000;
      case sqMIDIHasInputClock:      return 0;
      default:
         return success(false);
      }
   } else {
      switch(whichParameter) {
      case sqMIDIInstalled:
      case sqMIDIVersion:
      case sqMIDIHasBuffer:
      case sqMIDIHasDurs:
      case sqMIDICanSetClock:
      case sqMIDICanUseSemaphore:
         return success(false);
         break;
      case sqMIDIEchoOn:
         break;
      case sqMIDIUseControllerCache:
         break;
      case sqMIDIEventsAvailable:
         return success(false);
         break;
      case sqMIDIFlushDriver:
         break;
      case sqMIDIClockTicksPerSec:
         return success(false);
         break;
      default:
         return success(false);
      }
   }
   return 1;
}

int sqMIDIParameterSet(int whichParameter, int newValue)
{
   return sqMIDIParameter(whichParameter, 1, newValue);
}

int sqMIDIParameterGet(int whichParameter)
{
   return sqMIDIParameter(whichParameter, 0, 0);
}

#endif /* NO_MIDI */
